# ==============================================================================
"""GravityGrid : a generic rectangular grid with gravity"""
# ==============================================================================
__author__  = "Christophe Schlick"
__version__ = "1.0" # basic implementation without checking nor encapsulation
__date__    = "2023-10-15"
__usage__   = ""
# ------------------------------------------------------------------------------
from ezCLI import testcode, grid as make_grid
# ------------------------------------------------------------------------------
class GravityGrid(object):
  """a generic rectangular grid with gravity"""
  # ----------------------------------------------------------------------------
  def __init__(self, row=1, col=1, none=None):
    """initialize 'self' with provided arguments"""
    self.row, self.col, self.none, self.size = row, col, none, row*col
    self.clear() # set all grid items to 'none' value 
  # ----------------------------------------------------------------------------
  def __repr__(self):
    """return object representation of 'self'"""
    return f"GravityGrid(row={self.row}, col={self.col}, none={self.none!r})"
  # ----------------------------------------------------------------------------
  def __eq__(self, peer):
    """test equality between 'self' and 'peer'"""
    return repr(self) == repr(peer) and self() == peer()
  # ----------------------------------------------------------------------------
  def __str__(self):
    """return a grid representation of 'self' (using 'make_grid' function)"""
    return make_grid(self._grid)
  # ----------------------------------------------------------------------------
  def __call__(self, *items):
    """get grid as a 2D list or set grid from a 1D tuple of items"""
    if len(items): # 'items' contains a tuple of items to be inserted into grid
      for n, item in enumerate(items[:self.size]): # loop over items
        self._grid[n // self.col][n % self.col] = item
    return self._grid # return items as a 2D list
   # ---------------------------------------------------------------------------
  def __len__(self):
    """return length of 'self' (= number of non-empty items)"""
    items = sum(self._grid, []) # convert grid from 2D list to 1D list
    return len([item for item in items if item != self.none]) # count non-empty
  # ----------------------------------------------------------------------------
  def __getitem__(self, pos):
    """x.__getitem__(pos) <==> x[pos] where 'pos' is a (row,col) tuple"""
    row, col = pos; return self._grid[row][col] # get item at position 'pos'
  # ----------------------------------------------------------------------------
  def __setitem__(self, pos, item):
    """x.__setitem__(pos, item) <==> x[pos] = item"""
    row, col = pos; self._grid[row][col] = item # set item at position 'pos'
  # ----------------------------------------------------------------------------
  def __delitem__(self, pos):
    """x.__delitem__(pos) <==> del x[pos] where 'pos' is a (row,col) tuple"""
    self[pos] = self.none # delete item means set it to 'self.none'
  # ----------------------------------------------------------------------------
  def clear(self):
    """reset all the items to 'self.none'"""
    self._grid = [[self.none for col in range(self.col)]
      for row in range(self.row)]; return self._grid
  # ----------------------------------------------------------------------------
  def gravity(self):
    """apply gravity rules to current grid"""
    for col in range(self.col): # loop over cols and apply gravity to items
      # put all non-empty items at the bottom of each column 
      items = [self._grid[row][col] for row in range(self.row)]
      items = [item for item in items if item != self.none]
      items = [self.none] * (self.row - len(items)) + items
      for row in range(self.row): self._grid[row][col] = items[row]
# ==============================================================================
if __name__ == "__main__": # test code for class "GravityGrid"
  code = r'''
a = GravityGrid(3, 6, 0)
a, a.row, a.col, a.none, len(a)
a()
str(a)
a(1,2,3,4,5,6,7,8,9) # set initial items
len(a)
str(a)
a.gravity() # apply gravity
str(a)
a[0,1] = a[0,3] = a[0,4] = a[0,5] = 1 # add 4 items on top row
a[2,1] = a[2,3] = 0 # remove 2 items on bottom row
del a[2,0] # alternative way to remove item
len(a)
str(a)
a.gravity() # apply gravity
str(a)
a.clear()
len(a)
str(a)
'''; testcode(code)
# ==============================================================================
